from time import sleep
import traceback

from praw.exceptions import APIException
from prawcore.exceptions import BadRequest, ResponseException, RequestException
from requests.exceptions import ConnectionError

# from ..models import Logger

def retry(action):
    def actionWithRetry(*args, **kwargs):
        '''Perform the given action. If an exception is raised, retry every ten seconds

        @param throwOnPayloadTooLarge boolean If a 413 error is caught, throw it instead of retrying (default False)
        Return the return value of the action, if any
        '''

        failCount = 0

        while True:
            try:
                result = action(*args, **kwargs)

                # if failCount > 0:
                #     Logger.warn("Action succeeded after previously failing", {
                #         "action": action.__name__,
                #         "failCount": failCount,
                #     })

                return result

            except BadRequest as e:
                # Don't keep trying if we get a bad request
                # This is likely caused by deleted accounts so we can safely ignore them
                # Logger.warn("Action failed with HTTP status 400. Returning...", {
                #     "action": action.__name__,
                #     "traceback": traceback.format_exc(),
                #     "responseText": e.response.text,
                #     "responseUrl": e.response.url,
                #     "requestUrl": e.response.request.url,
                #     "requestBody": e.response.request.body,
                # })
                return

            except APIException as e:
                if e.error_type == "TOO_LONG" and kwargs.get('throwOnPayloadTooLarge', False):
                    # Logger.warn("Action failed with HTTP status 413, raising", { "action": action.__name__ })
                    raise e

                failCount += 1

                if failCount == 1:
                    pass
                    # Logger.error(traceback.format_exc())

                sleep(10)
                continue

            except (ConnectionError, ResponseException, RequestException) as e:
                failCount += 1

                if failCount == 1:
                    pass
                    # Logger.error(traceback.format_exc())

                sleep(10)
                continue

    return actionWithRetry



from .config import pg
from .Utils.utils import randomColor, getRoundPrefix
import time

def submitRound(r):
    submission = pg.submit(title=f'{getRoundPrefix()} {r["title"]}',
                           url=r['url'].strip())

    message = r.get('message')
    if message is not None:
        time.sleep(15)
        submission.reply(message)

        print(f'{randomColor()}Message posted to thread: {message}')

    return submission


from .config import pg, username, hints
from .posterino import submitRound
from .hosterino import checkAnswers, checkHints
from .Utils.utils import waitForApproval, approved, postDelay, randomColor
from .Loader.loader import getRound
from sty import fg
from importlib.metadata import version
from update_checker import UpdateChecker
from threading import Thread
import time

v = version('botterino')
checker = UpdateChecker()
result = checker.check('botterino', v)
if result:
    print(f'{fg.yellow}{result}')
    print(f'{fg.yellow}run "pip install --upgrade botterino" to update')

def checkType(r):
    types = []
    if 'tolerance' in r and 'answer' in r:
        types.append('coordinates')
    if 'tolerances' in r and 'answers' in r:
        types.append('multiple coordinates')
    if 'text' in r and 'similarity' in r:
        types.append('text match')
    if 'manual' in r:
        types.append('x wrong guesses with manual correct')
    if not types:
        return 'no automatic replies'
    if 'manual' not in r:
        types.append('automatic')
    return ','.join(types)

def main():
    while True:
        print(f'{fg.yellow}Waiting for {username} to win a round... 🐌')
        waitForApproval()
        print(f'{fg.blue}Congrats on a well deserved win {username}! ⭐')
        r = getRound()
        while not r:
            print(f'{fg.red}No rounds in round file! checking again in 10s')
            time.sleep(10)
            r = getRound()
        submission = submitRound(r)
        print(f'{randomColor()}Your round was posted to https://reddit.com{submission.permalink}')
        print(f'{fg.magenta}Round \'{r["title"]}\' posted in {postDelay()}s')
        print(f'{fg.cyan}Checking Answers: {checkType(r)}...')
        H = hints if not r.get('hints') else r.get('hints')
        CheckAnswers = Thread(target=checkAnswers, args=(r, submission))
        CheckHints = Thread(target=checkHints, args=(H, submission))
        CheckAnswers.start()
        CheckHints.start()
        CheckAnswers.join()
        CheckHints.join()
        while approved():
            continue
        after = r.get('after')
        if after:
            submission.reply(after)
            print(f'{randomColor()}Posted your message after the round: {after}')

if __name__ == 'botterino.botterino':
    main()




import praw
import os
import sys
from . import botfiles, correctMessage, incorrectMessage, hints
from sty import fg
import random
import webbrowser
import socket
from prawcore import exceptions


def receive_connection():
    """
    Wait for and then return a connected socket..
    Opens a TCP connection on port 8080, and waits for a single client.
    """
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(('localhost', 8080))
    server.listen(1)
    client = server.accept()[0]
    server.close()
    return client


def send_message(client, message):
    """
    Send message to client and close the connection.
    """
    client.send('HTTP/1.1 200 OK\r\n\r\n{}'.format(message).encode('utf-8'))
    client.close()


if sys.platform == 'win32':
    os.system('color')

debug = False
roundfile = botfiles.rounds
archivefile = botfiles.archive
cwd = os.getcwd()
try:
    os.chdir(botfiles.botconfig)
    reddit = praw.Reddit('botterino')
    print(f'{fg.green}Successfully logged into reddit as {reddit.user.me()}')
except exceptions.OAuthException:
    try:
        print(f'{fg.yellow}Unable to login with username/password. If your account has 2fa continue in browser. Otherwise check {botfiles.prawconfig}')
        os.chdir(botfiles.botconfig)
        reddit = praw.Reddit(
            'botterino', redirect_uri='http://localhost:8080')
        state = str(random.randint(0, 65000))
        scopes = ['identity', 'history', 'read',
                  'edit', 'submit', 'privatemessages']
        url = reddit.auth.url(scopes, state, 'permanent')
        print(
            'A window will be opened in the browser to complete the login process to reddit.')
        webbrowser.open(url)

        client = receive_connection()
        data = client.recv(1024).decode('utf-8')
        param_tokens = data.split(' ', 2)[1].split('?', 1)[1].split('&')
        params = {key: value for (key, value) in [token.split('=')
                                                  for token in param_tokens]}

        if state != params['state']:
            send_message(client, 'State mismatch. Expected: {} Received: {}'.format(
                state, params['state']))
        elif 'error' in params:
            send_message(client, params['error'])

        refresh_token = reddit.auth.authorize(params["code"])
        send_message(client, "Refresh token: {}".format(refresh_token))

        print(refresh_token)
    except Exception as e:
        print(f'{fg.red}Unable to login to reddit. Please check {botfiles.prawconfig}')
        print(f'{fg.red}{e}')
except Exception as e:
    print(f'{fg.red}Unable to login to reddit. Please check {botfiles.prawconfig}')
    print(f'{fg.red}{e}')
finally:
    os.chdir(cwd)

pg = reddit.subreddit('itoxtestingfacility' if debug else 'picturegame')

username = str(reddit.user.me())

donotreply = {
    'achievements-bot',
    username.lower(),
    'r-picturegame',
    'imreallycuriousbird',
}




from pathlib import Path
from ..config import roundfile, archivefile
import ruamel.yaml

yaml = ruamel.yaml.YAML()
yaml.preserve_quotes = True
yaml.allow_duplicate_keys = True

def dump(data, file):
    with open(file, 'w', encoding='utf-8') as f:
        yaml.dump(data, f)

def load(file):
    with open(file, 'r', encoding='utf-8') as f:
        return yaml.load(f)

def getRound():
    x = load(roundfile)
    if not x:
        return None
    k = next(iter(x))
    top = x.pop(k)
    if x:
        dump(x, roundfile)
    else:
        open(roundfile, 'w', encoding='utf-8').close()
    y = load(archivefile) or {}
    y[k] = top
    dump(y, archivefile)
    return top


import os
import json
from pathlib import Path
from configparser import ConfigParser

class files():
    def __init__(self):
        self.home = os.path.expanduser('~')
        self.botconfig = os.path.join(self.home, 'botterino-config')
        self.configfile = os.path.join(self.botconfig, 'config.ini')
        self.roundsdir = os.path.join(self.botconfig, 'rounds')
        self.rounds = os.path.join(self.roundsdir, 'rounds.yaml')
        self.archive = os.path.join(self.roundsdir, 'archive.yaml')
        self.prawconfig = os.path.join(self.botconfig, 'praw.ini')
        self.hintfile = os.path.join(self.botconfig, 'hints.txt')

botfiles = files()
dirs = [botfiles.botconfig, botfiles.roundsdir]
files = [botfiles.rounds, botfiles.archive, botfiles.prawconfig, botfiles.hintfile]
for d in dirs:
    Path(d).mkdir(parents=True, exist_ok=True)
for f in files:
    with open(f, 'a+'):
        pass

if not os.path.exists(botfiles.configfile):
    parser = ConfigParser()
    parser.add_section('config')
    parser.set('config', 'correct_message', '+correct')
    parser.set('config', 'incorrect_message', '❌')
    parser.set('config', 'hints', '[25,45]')
    with open(botfiles.configfile, 'w', encoding='utf-8') as f:
        parser.write(f)

parser = ConfigParser()
parser.read(botfiles.configfile, encoding='utf-8')
correctMessage, incorrectMessage = parser['config']['correct_message'], parser[
    'config']['incorrect_message']
hints = json.loads(parser['config'].get('hints', '[]'))


from ..config import pg, username, debug, reddit, donotreply
from ..RedditPoller.RedditPoller import RedditPoller, CommentWrapper
from ..RedditPoller.Retry import retry
from geopy.point import Point
from geopy.distance import distance
from sty import fg
from random import randrange
from halo import Halo
import re
import requests
import json
import time
import warnings


decimal = re.compile("""([-+]?\d{1,2}[.]\d+),\s*([-+]?\d{1,3}[.]\d+)""")

# google maps decimal or dms
decimal_or_DMS = re.compile(
    """(?:((?:[-+]?\d{1,2}[.]\d+),\s*(?:[-+]?\d{1,3}[.]\d+))|(\d{1,3}°\d{1,3}'\d{1,3}\.\d\"[N|S]\s\d{1,3}°\d{1,3}'\d{1,3}\.\d\"[E|W]))"""
)

# google earth formats, other formats
everything_else = re.compile(
    """(^| )(-?\d{1,2}(\.\d+)?(?=\s*,?\s*)[\s,]+-?\d{1,3}(\.\d+)?|\d{1,2}(\.\d+°|°(\d{1,2}(\.\d+'|'(\d{1,2}(\.\d+)?\")?))?)[NS](?=\s*,?\s*)[\s,]+\d{1,3}(\.\d+°|°(\d{1,2}(\.\d+'|'(\d{1,2}(\.\d+)?\")?))?)[EW])"""
)

MAPS_URL = 'https://maps.google.com/maps?t=k&q=loc:{},{}'

badColors = [0, 15, 16] + list(range(231,256))

def randomColorWithAuthor(author):
    color = abs(hash(author)) % 256
    while color in badColors:
        author += '*'
        color = abs(hash(author)) % 256
    return fg(color)

def randomColor():
    color = randrange(256)
    while color in badColors:
        color = randrange(256)
    return fg(color)

def postDelay():
    if debug:
        return -1

    r = requests.get('https://api.picturegame.co/current')
    data = json.loads(r.content)
    tries = 1

    while data['round']['hostName'].lower() != username.lower():
        tries += 1
        if tries > 5:
            return -1
        r = requests.get('https://api.picturegame.co/current')
        data = json.loads(r.content)
        time.sleep(5)

    return data['round'].get('postDelay', -1)

def getRoundPrefix():
    r = requests.get('https://api.picturegame.co/current')
    roundnum = int(json.loads(r.content)['round']['roundNumber']) + 1
    return f'[Round {roundnum}]'

def approved():
    c = next(iter(pg.contributor()))
    return c and c.name.lower() == username.lower()

@retry
@Halo(spinner='dots', color='yellow')
def waitForApproval():
    while not approved():
        continue

def getDistance(guess, answer):
    match = re.search(decimal_or_DMS, guess)
    if not match:
        match = re.search(everything_else, guess)
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            coord = Point(match[0]) if match else Point(guess)
    except ValueError as v:
        return
    try:
        return distance(coord, answer).m, coord
    except Exception as e:
        return


def getComments(submission):
    rp = RedditPoller(CommentWrapper(pg.comments, reddit.inbox.all))
    comments = rp.getLatest()
    # flush old comments
    while next(comments):
        continue

    for c in comments:
        if hasattr(c, 'submission') and hasattr(c, 'author') and c.author and c.submission:
            if c.author.name.lower() == username.lower() and '+correct' in c.body and not c.is_root:
                return
            if not c.is_root:
                continue
            if c.author.name.lower() in donotreply or c.submission != submission:
                continue
            yield c



from geopy.distance import distance
from geopy.point import Point
from .config import donotreply, correctMessage, incorrectMessage, reddit, username, pg, botfiles
from itertools import permutations
import re
from sty import fg
from .Utils.utils import approved, decimal, getComments, getDistance, randomColor, randomColorWithAuthor, MAPS_URL
from difflib import SequenceMatcher
import time

def withinTolerance(guess, answer, tolerance):
    return distance(guess, answer).m <= tolerance

def checkMultipleCoordinates(guess, answers, tolerances):
    guesser = guess.author.name
    answers = [Point(a) for a in answers]
    match = re.findall(decimal, guess.body)
    if len(match) != len(answers):
        # TODO print a message here
        print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess {guess.body} was incorrect')
        return False
    try:
        points = [Point(f'{lat},{lon}') for lat, lon in match]
        points = permutations(points)
    except Exception as e:
        print(f'{randomColor()}Something happened: ', e, 'it probably does not matter')
        return False

    results = [[withinTolerance(p, a, t) for p, a, t in zip(ps, answers, tolerances)] for ps in points]
    results = [all(r) for r in results]
    result = any(results)
    if not result:
        print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess {guess.body} was incorrect')
    return result

def checkCoordinates(guess, answer, tolerance):
    guesser = guess.author.name
    answer = Point(answer)
    errorAndPoint = getDistance(guess.body, answer)
    error, point = errorAndPoint if errorAndPoint else (None, None)
    if error is None:
        print(f"{randomColorWithAuthor(guesser)}Could not find a coordinate in guess '{guess.body}' by {guesser}")
        return 'ignore'
    error = round(error, 2)
    mapslink = MAPS_URL.format(point.latitude, point.longitude)
    print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess {mapslink} was {error} meters off')
    return error <= tolerance

def checkText(guess, answer, tolerance, ignorecase):
    guesser = guess.author.name
    text = guess.body.strip().replace('\\', '')

    if ignorecase:
        text,answer = text.lower(), answer.lower()

    similarity = SequenceMatcher(None, text, answer).ratio()
    print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess was {round(similarity * 100, 3)}% similar to the correct answer')
    return similarity >= tolerance

def postHint(submission, time):
    with open(botfiles.hintfile, 'r') as F:
        hintText = F.read()
    if not hintText:
        print(f'{fg.yellow}Skipping {time}m hint: hints.txt is empty')
        return
    hint = submission.reply(f'Hint({time}m): {hintText}')
    print(f'{fg.green}Posted hint ({time}m) to https://reddit.com{hint.permalink}')
    open(botfiles.hintfile, 'w').close()

def checkHints(hints, submission):
    hints = list(hints)
    hints += [60, 120, 180, 240]
    hints = sorted(list(set(hints)))
    while hints and approved():
        top = hints[0]
        duration = int(time.time() - submission.created_utc)
        duration = duration//60
        if duration >= top:
            postHint(submission, duration)
            hints.pop(0)
        time.sleep(10)
        pass

def checkAnswers(r, submission):
    tolerance, manual, text, answer, tolerances, answers, similarity, ignorecase, hints, = r.get(
        'tolerance'), r.get('manual'), r.get('text'), r.get(
            'answer'), r.get('tolerances'), r.get('answers'), r.get(
                'similarity'), r.get('ignorecase'), r.get('hints', [])

    if tolerance is None and tolerances is None and text is None:
        return

    for c in getComments(submission):
        result = True
        if tolerance:
            tolerance = float(tolerance)
            r = checkCoordinates(c, answer, tolerance)
            if r == 'ignore':
                continue
            result = result and r
        elif tolerances:
            tolerances = [float(t) for t in r['tolerances']]
            if len(answers) != len(tolerances):
                print('{fg.red}Refusing to check answers, number of tolerances must equal number of answers.')
            result = result and checkMultipleCoordinates(c, answers, tolerances)

        if text and similarity is None:
            continue

        if ignorecase is None:
            ignorecase = True

        if text:
            result = result and checkText(c, text, similarity, ignorecase)

        if not result:
            c.reply(incorrectMessage)
        if result:
            if manual:
                print(f"{randomColor()}Guess '{c.body}' looks correct, but you will have to check it out.")
            else:
                plusCorrect = c.reply(correctMessage)
                guesser = c.author.name
                print(
                    f'{randomColorWithAuthor(guesser)}Corrected {guesser} in {plusCorrect.created_utc - c.created_utc}s')
                break
